/*
----------------------------------------------------------------------
volume.c
----------------------------------------------------------------------
This program reads from stdin a triangulated surface of a simple
polyhedron, and computes its volume. The input format is as follows:

   n   ( # of vertices )
   x0 y0 z0
   x1 y1 z1
   ...
   xn-1 yn-1 zn-1
   F   ( # of faces/triangles )
   i0 i1 i2
   ...

The vertex coordinates are integers.  The triangle faces are each
represented by three vertex indices.  All faces are assumed oriented
counterclockwise from the outside; this important condition is not
checked in the code.  See the sample input files i.cube & i.nonconvex.
Compile: make.
Run: volume < i.cube.

Written by Joseph O'Rourke and Min Xu, June 1997.
Used in the textbook, "Computational Geometry in C" (Second Edition).
Questions to orourke@cs.smith.edu.
--------------------------------------------------------------------
This code is Copyright 1997 by Joseph O'Rourke.  It may be freely
redistributed in its entirety provided that this copyright notice is
not removed.
--------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
#define	X	0
#define	Y	1
#define Z       2

#define DIM     3               /* Dimension of points */
typedef int     tPointi[DIM];   /* type integer point */
#define PMAX    10000           /* Max # of pts in polygon */
   
int  ReadVertex( tPointi Vertices[] );
int  ReadFace( tPointi Faces[] );
int  Volume6( tPointi a, tPointi b, tPointi c );

main()
{
  int n, F, i, vol = 0;
  tPointi Vertices[PMAX];
  tPointi Faces[PMAX];

  n = ReadVertex( Vertices );
  F = ReadFace( Faces );

  /* Euler's formula for a triangulated surface of a simple polyhedron. */
  if ( F != 2 * n - 4 )
    printf( "Not a simple polyhedron!\n" ),
    exit(1);

  for ( i = 0; i < F; i++ )
    vol += Volume6( Vertices[ Faces[i][0] ], 
                    Vertices[ Faces[i][1] ], 
                    Vertices[ Faces[i][2] ] );
  printf("Volume = %g\n", vol/6.0 );
}

/* Reads in the number and coordinates of the vertices of a polyhedron
   from stdin, and returns n, the number of vertices. */
int ReadVertex( tPointi Vertices[] )
{
  int	i, n;
  
  do {
    scanf( "%d", &n );
    if ( n <= PMAX )
      break;
    printf("Error in read_vertex:  too many points; max is %d\n", PMAX);
  }
  while ( 1 );

  printf( "Polyhedron Vertices:\n" );
  printf( "  i:   x   y   z\n");
  for ( i = 0; i < n; i++ ) {
    scanf( "%d %d %d", &Vertices[i][X], &Vertices[i][Y], &Vertices[i][Z] );
    printf( "%3d:%4d%4d%4d\n", i, Vertices[i][X], Vertices[i][Y], Vertices[i][Z] );
  }
  printf("n = %3d vertices read\n",n);
  putchar('\n');

  return n;
}

/* Reads in the number of faces of the polyhedron and their indices from stdin,
   and returns the number n. */
int ReadFace( tPointi Faces[] )
{
  int	i, n;
  
  do {
    scanf( "%d", &n );
    if ( n <= PMAX )
      break;
    printf("Error in read_vertex:  too many points; max is %d\n", PMAX);
  }
  while ( 1 );

  printf( "Faces:\n" );
  printf( "  i0  i1  i2\n");
  for ( i = 0; i < n; i++ ) {
    scanf( "%d %d %d", &Faces[i][0], &Faces[i][1], &Faces[i][2] );
    printf( "%3d%4d%4d\n", Faces[i][0], Faces[i][1], Faces[i][2] );
  }
  printf("n = %3d faces read\n",n);
  putchar('\n');

  return n;
}

/*-------------------------------------------------------------------------
Volume6 returns six times the volume of the tetrahedron determined by abc 
and the origin d.  Volume6 is positive iff d is on the negative side of abc,
where the positive side is determined by the rh-rule.  So the volume 
is positive if the ccw normal to abc points outside the tetrahedron.
This code is taken from chull.c; see "Computational Geometry in C."
--------------------------------------------------------------------------*/
int 	Volume6( tPointi a, tPointi b, tPointi c )
{ 
	double	vol;
	double 	ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz;
        double	bxdx, bydy, bzdz, cxdx, cydy, czdz;

	ax = a[X];
	ay = a[Y];
	az = a[Z];
	bx = b[X];
	by = b[Y];
	bz = b[Z];
	cx = c[X]; 
	cy = c[Y];
	cz = c[Z];
        /* This could be simplified because point d is the origin. */
	dx = 0;
	dy = 0;
	dz = 0;

	bxdx=bx-dx;
        bydy=by-dy;
        bzdz=bz-dz;
        cxdx=cx-dx;
        cydy=cy-dy;
        czdz=cz-dz;
        vol =    (az-dz)*(bxdx*cydy-bydy*cxdx)
               + (ay-dy)*(bzdz*cxdx-bxdx*czdz)
               + (ax-dx)*(bydy*czdz-bzdz*cydy);
	return (int) vol;
}

